
from django.test import TestCase
from django_dag.models import *


# Test classes

class ConcreteNode(node_factory('ConcreteEdge')):
    """
    Test node, adds just one field
    """
    name = models.CharField(max_length = 32)

    class Meta:
        app_label = 'django_dag'

class ConcreteEdge(edge_factory('ConcreteNode', concrete = False)):
    """
    Test edge, adds just one field
    """
    name = models.CharField(max_length = 32, blank = True, null = True)

    class Meta:
        app_label = 'django_dag'


class DagTestCase(TestCase):

    def setUp(self):
        for i in range(1,11):
            ConcreteNode(name = "%s" % i).save()


    def test_01_objects_were_created(self):
        for i in range(1,11):
            self.assertEqual(ConcreteNode.objects.get(name="%s" % i).name, "%s" % i)


    def test_02_dag(self):
        # Get nodes
        for i in range(1,11):
            globals()["p%s" % i] = ConcreteNode.objects.get(pk=i)

        # Creates a DAG
        #from IPython import embed; embed()
        p1.add_child(p5)
        p5.add_child(p7)

        tree = p1.descendants_tree()
        # {<ConcreteNode: # 5>: {<ConcreteNode: # 7>: {}}}
        self.assertEqual(tree.keys()[0], p5)
        self.assertEqual(tree.values()[0].keys()[0], p7)
        self.assertEqual(tree[tree.keys()[0]].values()[0], {})

        l=[p.pk for p in p1.descendants_set()]
        l.sort()
        self.assertEqual(l, [5, 7])

        p1.add_child(p6)
        p2.add_child(p6)
        p3.add_child(p7)
        p6.add_child(p7)
        p6.add_child(p8)
        l=[p.pk for p in p2.descendants_set()]
        l.sort()
        self.assertEqual(l, [6, 7, 8])

        # ValidationError: [u'The object is a descendant.']
        self.assertRaises(ValidationError, p2.add_child, p8)

        try:
            p2.add_child(p8)
        except ValidationError, e:
            self.assertEqual(unicode(e[0]), u'The object is a descendant.')

        # Checks that p8 was not added two times
        l=[p.pk for p in p2.descendants_set()]
        l.sort()
        self.assertEqual(l, [6, 7, 8])

        p6.add_parent(p4)
        p9.add_parent(p3)
        p9.add_parent(p6)
        self.assertRaises(ValidationError, p9.add_child, p2)
        try:
            p9.add_child(p2)
        except ValidationError, e:
            self.assertEqual(unicode(e[0]), u'The object is an ancestor.')

        self.assertEqual(str(p1.descendants_tree()), """{<ConcreteNode: # 5>: {<ConcreteNode: # 7>: {}}, <ConcreteNode: # 6>: {<ConcreteNode: # 8>: {}, <ConcreteNode: # 9>: {}, <ConcreteNode: # 7>: {}}}""")
        l=[p.pk for p in p1.descendants_set()]
        l.sort()
        self.assertEqual(l, [5, 6, 7, 8, 9])
        self.assertEqual(p1.distance(p8), 2)

        #Test additional fields for edge
        p9.add_child(p10, name = 'test_name')
        self.assertEqual(p9.children.through.objects.filter(child=p10)[0].name, u'test_name')

        self.assertEqual(str(p1.path(p7)), '[<ConcreteNode: # 6>, <ConcreteNode: # 7>]')
        self.assertEqual(str(p1.path(p10)), '[<ConcreteNode: # 6>, <ConcreteNode: # 9>, <ConcreteNode: # 10>]')
        self.assertEqual(p1.distance(p7), 2)

        self.assertEqual(str(p1.get_leaves()), 'set([<ConcreteNode: # 8>, <ConcreteNode: # 10>, <ConcreteNode: # 7>])')
        self.assertEqual(str(p8.get_roots()), 'set([<ConcreteNode: # 1>, <ConcreteNode: # 2>, <ConcreteNode: # 4>])')

        self.assertTrue(p1.is_root())
        self.assertFalse(p1.is_leaf())
        self.assertFalse(p10.is_root())
        self.assertTrue(p10.is_leaf())
        self.assertFalse(p6.is_leaf())
        self.assertFalse(p6.is_root())

        self.assertRaises(ValidationError, p6.add_child, p6)
        try:
            p6.add_child(p6)
        except ValidationError, e:
            self.assertEqual(unicode(e[0]), u'Self links are not allowed.')

        # Remove a link and test island
        p10.remove_parent(p9)
        self.assertFalse(p10 in p9.descendants_set())
        self.assertTrue(p10.is_island())

        self.assertEqual(str(p6.ancestors_set()), 'set([<ConcreteNode: # 1>, <ConcreteNode: # 2>, <ConcreteNode: # 4>])')

        p1.remove_child(p6)
        self.assertEqual(str(p6.ancestors_set()), 'set([<ConcreteNode: # 2>, <ConcreteNode: # 4>])')

        self.assertFalse(p1 in p6.ancestors_set())

        # Testing the view
        from django.shortcuts import render_to_response
        response = str(render_to_response('tree.html', { 'dag_list': ConcreteNode.objects.all()}))
        self.assertEqual(response, """Content-Type: text/html; charset=utf-8\r\n\r\n# 1\nDescendants:\n# 5\n# 7\n\n\n\n# 2\nDescendants:\n# 6\n# 8\n\n# 9\n\n# 7\n\n\n\n# 3\nDescendants:\n# 9\n\n# 7\n\n\n# 4\nDescendants:\n# 6\n# 8\n\n# 9\n\n# 7\n\n\n\n# 5\n\n# 6\n\n# 7\n\nAncestors:\n# 3\n\n# 5\n# 1\n\n\n# 6\n# 2\n\n# 4\n\n\n# 8\n\nAncestors:\n# 6\n# 2\n\n# 4\n\n\n# 9\n\nAncestors:\n# 3\n\n# 6\n# 2\n\n# 4\n\n\n# 10\n\n\n""")
        



"""
A class to model hierarchies of objects following
Directed Acyclic Graph structure.

Some ideas stolen from: from https://github.com/stdbrouw/django-treebeard-dag

"""


from django.db import models
from django.core.exceptions import ValidationError


class NodeNotReachableException (Exception):
    """
    Exception for node distance and path
    """
    pass


class NodeBase(object):
    """
    Main node abstract model
    """

    class Meta:
        ordering = ('-id',)

    def __unicode__(self):
        return "# %s" % self.pk

    def __str__(self):
        return self.__unicode__()

    def add_child(self, descendant, **kwargs):
        """
        Adds a child
        """
        args = kwargs
        args.update({'parent' : self, 'child' : descendant })
        disable_check = args.pop('disable_circular_check', False)
        cls = self.children.through(**kwargs)
        return cls.save(disable_circular_check=disable_check)


    def add_parent(self, parent, *args, **kwargs):
        """
        Adds a parent
        """
        return parent.add_child(self, **kwargs)

    def remove_child(self, descendant):
        """
        Removes a child
        """
        self.children.through.objects.get(parent = self, child = descendant).delete()

    def remove_parent(self, parent):
        """
        Removes a parent
        """
        parent.children.through.objects.get(parent = parent, child = self).delete()

    def parents(self):
        """
        Returns all elements which have 'self' as a direct descendant
        """
        return self.__class__.objects.filter(children = self)

    def descendants_tree(self):
        """
        Returns a tree-like structure with progeny
        """
        tree = {}
        for f in self.children.all():
            tree[f] = f.descendants_tree()
        return tree

    def ancestors_tree(self):
        """
        Returns a tree-like structure with ancestors
        """
        tree = {}
        for f in self.parents():
            tree[f] = f.ancestors_tree()
        return tree

    def descendants_set(self):
        """
        Returns a set of descendants
        """
        res = set()
        for f in self.children.all():
            res.add(f)
            res.update(f.descendants_set())
        return res

    def ancestors_set(self):
        """
        Returns a set of ancestors
        """
        res = set()
        for f in self.parents():
            res.add(f)
            res.update(f.ancestors_set())
        return res

    def distance(self, target):
        """
        Returns the shortest hops count to the target vertex
        """
        return len(self.path(target))

    def path(self, target):
        """
        Returns the shortest path
        """
        if self == target:
            return []
        if target in self.children.all():
            return [target]
        if target in self.descendants_set():
            path = None
            for d in self.children.all():
                try:
                    desc_path = d.path(target)
                    if not path or len(desc_path) < path:
                        path = [d] + desc_path
                except NodeNotReachableException:
                    pass
        else:
            raise NodeNotReachableException
        return path

    def is_root(self):
        """
        Check if has children and not ancestors
        """
        return bool(self.children.exists() and not self._parents.exists())

    def is_leaf(self):
        """
        Check if has ancestors and not children
        """
        return bool(self._parents.exists() and not self.children.exists())

    def is_island(self):
        """
        Check if has no ancestors nor children
        """
        return bool(not self.children.exists() and not self._parents.exists())

    def _get_roots(self, at):
        """
        Works on objects: no queries
        """
        if not at:
          return set([self])
        roots = set()
        for a2 in at:
            roots.update(a2._get_roots(at[a2]))
        return roots

    def get_roots(self):
        """
        Returns roots nodes, if any
        """
        at =  self.ancestors_tree()
        roots = set()
        for a in at:
            roots.update(a._get_roots(at[a]))
        return roots

    def _get_leaves(self, dt):
        """
        Works on objects: no queries
        """
        if not dt:
          return set([self])
        leaves = set()
        for d2 in dt:
            leaves.update(d2._get_leaves(dt[d2]))
        return leaves

    def get_leaves(self):
        """
        Returns leaves nodes, if any
        """
        dt =  self.descendants_tree()
        leaves = set()
        for d in dt:
            leaves.update(d._get_leaves(dt[d]))
        return leaves


    @staticmethod
    def circular_checker(parent, child):
        """
        Checks that the object is not an ancestor or a descendant,
        avoid self links
        """
        if parent == child:
            raise ValidationError('Self links are not allowed.')
        if child in parent.ancestors_set():
            raise ValidationError('The object is an ancestor.')
        if child in parent.descendants_set():
            raise ValidationError('The object is a descendant.')



def edge_factory(node_model, child_to_field = "id", parent_to_field = "id", concrete = True, base_model = models.Model):
    """
    Dag Edge factory
    """
    if isinstance(node_model, str):
        try:
            node_model_name = node_model.split('.')[1]
        except IndexError:
            node_model_name = node_model
    else:
        node_model_name = node_model._meta.module_name

    class Edge(base_model):
        class Meta:
            abstract = not concrete

        parent = models.ForeignKey(node_model, related_name = "%s_child" % node_model_name, to_field = parent_to_field)
        child = models.ForeignKey(node_model, related_name = "%s_parent" % node_model_name, to_field = child_to_field)

        def __unicode__(self):
            return "%s is child of %s" % (self.child, self.parent)

        def save(self, *args, **kwargs):
            if not kwargs.pop('disable_circular_check', False):
                self.parent.__class__.circular_checker(self.parent, self.child)
            super(Edge, self).save(*args, **kwargs) # Call the "real" save() method.

    return Edge

def node_factory(edge_model, children_null = True, base_model = models.Model):
    """
    Dag Node factory
    """
    class Node(base_model, NodeBase):
        class Meta:
            abstract        = True

        children  = models.ManyToManyField(
                'self',
                null        = children_null,
                blank       = children_null,
                symmetrical = False,
                through     = edge_model,
                related_name = '_parents') # NodeBase.parents() is a function

    return Node




#!/usr/bin/env python

import os
from distutils.core import setup

version = '1.3'

classifiers = [
    "Development Status :: 4 - Beta",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: GNU Affero General Public License v3",
    "Programming Language :: Python",
    "Operating System :: OS Independent",
    "Topic :: Software Development :: Libraries",
    "Topic :: Utilities",
    "Environment :: Web Environment",
    "Framework :: Django",
]

root_dir = os.path.dirname(__file__)
if not root_dir:
    root_dir = '.'
long_desc = open(root_dir + '/README').read()

setup(
    name='django-dag',
    version=version,
    url='https://github.com/elpaso/django-dag',
    author='Alessandro Pasotti',
    author_email='apasotti@gmail.com',
    license='GNU Affero General Public License v3',
    packages=['django_dag'],
    package_dir={'django_dag': 'django_dag'},
    #package_data={'dag': ['templates/admin/*.html']},
    description='Directed Acyclic Graph implementation for Django 1.6+',
    classifiers=classifiers,
    long_description=long_desc,
)



# dict recurse template tag for django
# from http://djangosnippets.org/snippets/1974/

from django import template

register = template.Library()

from django import template

register = template.Library()


class RecurseDictNode(template.Node):
    def __init__(self, var, nodeList):
        self.var = var
        self.nodeList = nodeList

    def __repr__(self):
        return '<RecurseDictNode>'

    def renderCallback(self, context, vals, level):
        if len(vals) == 0:
            return ''

        output = []

        if 'loop' in self.nodeList:
            output.append(self.nodeList['loop'].render(context))

        for k, v in vals:
            context.push()

            context['level'] = level
            context['key'] = k

            if 'value' in self.nodeList:
                output.append(self.nodeList['value'].render(context))

                if type(v) == list or type(v) == tuple:
                    child_items = [ (None, x) for x in v ]
                    output.append(self.renderCallback(context, child_items, level + 1))
                else:
                    try:
                        child_items = v.items()
                        output.append(self.renderCallback(context, child_items, level + 1))
                    except:
                        output.append(unicode(v))

            if 'endloop' in self.nodeList:
                output.append(self.nodeList['endloop'].render(context))
            else:
                output.append(self.nodeList['endrecursedict'].render(context))

            context.pop()

        if 'endloop' in self.nodeList:
            output.append(self.nodeList['endrecursedict'].render(context))

        return ''.join(output)

    def render(self, context):
        vals = self.var.resolve(context).items()
        output = self.renderCallback(context, vals, 1)
        return output

def recursedict_tag(parser, token):
    bits = list(token.split_contents())
    if len(bits) != 2 and bits[0] != 'recursedict':
        raise template.TemplateSyntaxError, "Invalid tag syntax expected '{% recursedict [dictVar] %}'"

    var = parser.compile_filter(bits[1])
    nodeList = {}
    while len(nodeList) < 4:
        temp = parser.parse(('value','loop','endloop','endrecursedict'))
        tag = parser.tokens[0].contents
        nodeList[tag] = temp
        parser.delete_first_token()
        if tag == 'endrecursedict':
            break

    return RecurseDictNode(var, nodeList)

recursedict_tag = register.tag('recursedict', recursedict_tag)
